#  -*- coding: utf-8 -*-
# (c) 2018-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
"""
Run different animation scenarios/sequences.
As the ad-hoc interface is not finished yet, this is all we got. Enjoy!

The parameters `dtstart`, `dtuntil` and `interval` should explain themselves.

The parameter `mode` can have two values (defaulting to `window`):
- "window" will slide a window through the defined time range, with `interval` as window width
- "cumulative" will use a fixed start time and stepwise expand the end time by `interval`
"""
import logging
from datetime import datetime

from grafanimate.model import AnimationScenario, AnimationSequence, SequencingMode

logger = logging.getLogger(__name__)


def playdemo():
    """
    Run demo on `play.grafana.org`.

    Example::

        grafanimate --grafana-url=https://play.grafana.org/ --dashboard-uid=000000012 --scenario=playdemo
    """
    logger.info("Running scenario playdemo")

    return AnimationScenario(
        grafana_url="https://play.grafana.org/",
        dashboard_uid="000000012",
        steps=[
            AnimationSequence(
                dtstart=datetime(2021, 11, 14, 2, 0, 0),
                # Produce video with reasonable duration to explore different rendering options.
                dtuntil=datetime(2021, 11, 14, 4, 16, 36),
                # Produce very short video.
                # dtuntil=datetime(2021, 11, 14, 2, 16, 36),
                interval="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
        ],
    )


def playdemo_advanced():
    """
    Run demo on `play.grafana.org`.

    Example::

        grafanimate --grafana-url=https://play.grafana.org/ --dashboard-uid=000000012 --scenario=playdemo_advanced
    """
    logger.info("Running scenario playdemo")

    return AnimationScenario(
        grafana_url="https://play.grafana.org/",
        dashboard_uid="000000012",
        steps=[
            AnimationSequence(
                dtstart=datetime(2021, 11, 14, 2, 0, 0),
                dtuntil=datetime(2021, 11, 14, 2, 16, 36),
                interval="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                dtstart="2021-11-15T02:12:05Z",
                dtuntil="2021-11-15T02:37:36Z",
                interval="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
            AnimationSequence(
                dtstart=1637091011,
                dtuntil=1637091911,
                interval="5min",
                mode=SequencingMode.CUMULATIVE,
            ),
        ],
    )


def ldi_all():
    """
    Luftdaten.info, all
    """
    logger.info("Running scenario ldi_all")

    return [
        # LDI, ramp-up
        AnimationSequence(dtstart=datetime(2015, 10, 1), dtuntil=datetime(2017, 1, 1), interval="monthly"),
        # LDI, growth
        AnimationSequence(dtstart=datetime(2017, 1, 1), dtuntil=datetime.now(), interval="weekly"),
    ]


def ldi_with_gaps():
    """
    Luftdaten.info, growth
    """
    logger.info("Running scenario ldi_with_gaps")

    return [
        # LDI, ramp-up
        AnimationSequence(dtstart=datetime(2015, 10, 1), dtuntil=datetime(2017, 1, 1), interval="monthly"),
        # LDI, growth, with gap at 2018-04-29 - 2018-12-20
        # TODO: Detect empty data from datasource through Grafana Sidecar and skip respective images.
        AnimationSequence(dtstart=datetime(2017, 1, 1), dtuntil=datetime(2018, 6, 5), interval="weekly"),
        # LDI, until now
        AnimationSequence(dtstart=datetime(2018, 12, 20), dtuntil=datetime.now(), interval="weekly"),
    ]


def ldi_nye_2017_2018():
    """
    LDI, New Year's Eve 2018
    """
    logger.info("Running scenario ldi_nye_2017_2018")
    return AnimationSequence(
        dtstart=datetime(2017, 12, 31, 21, 0, 0), dtuntil=datetime(2018, 1, 1, 4, 0, 0), interval="10min"
    )


def ldi_nye_2018_2019():
    """
    LDI, New Year's Eve 2018/2019
    """
    logger.info("Running scenario ldi_nye_2018_2019")
    return [
        AnimationSequence(
            dtstart=datetime(2018, 12, 31, 15, 0, 0), dtuntil=datetime(2018, 12, 31, 20, 0, 0), interval="30min"
        ),
        AnimationSequence(
            dtstart=datetime(2018, 12, 31, 20, 0, 0), dtuntil=datetime(2018, 12, 31, 23, 0, 0), interval="10min"
        ),
        AnimationSequence(
            dtstart=datetime(2018, 12, 31, 23, 0, 0), dtuntil=datetime(2019, 1, 1, 1, 0, 0), interval="5min"
        ),
        AnimationSequence(
            dtstart=datetime(2019, 1, 1, 1, 0, 0), dtuntil=datetime(2019, 1, 1, 4, 0, 0), interval="10min"
        ),
        AnimationSequence(
            dtstart=datetime(2019, 1, 1, 4, 0, 0), dtuntil=datetime(2019, 1, 1, 9, 0, 0), interval="30min"
        ),
    ]


def cdc_maps():
    """
    DWD CDC, temperatur-sonne-and-niederschlag-karten
    """
    logger.info("Running scenario cdc_maps")
    return AnimationSequence(
        dtstart=datetime(2018, 3, 6, 5, 0, 0), dtuntil=datetime(2018, 3, 10, 23, 59, 59), interval="hourly"
    )

    # Short sequence, for debugging processes.
    # return AnimationSequence(dtstart=datetime(2018, 3, 6, 5, 0, 0), dtuntil=datetime(2018, 3, 6, 6, 59, 59), interval='hourly')


def uba_ldi_dwd_maps():
    """
    Labor: Studio / UBA/LDI/DWD-Studio [dev!]
    """
    logger.info("Running scenario uba_ldi_dwd_maps")
    return AnimationSequence(
        dtstart=datetime(2018, 10, 6, 5, 0, 0), dtuntil=datetime(2018, 10, 10, 23, 59, 59), interval="hourly"
    )


def ir_sensor_svg_pixmap():
    """
    dtstart: 2018-08-14 03:16:00
    dtuntil: 2018-08-14 03:16:36
    """
    logger.info("Running scenario ir_sensor_svg_pixmap")
    return AnimationSequence(
        dtstart=datetime(2018, 8, 14, 3, 16, 0), dtuntil=datetime(2018, 8, 14, 3, 16, 36), interval="secondly"
    )



#  -*- coding: utf-8 -*-
# (c) 2018 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
import time
from datetime import timedelta
from operator import attrgetter

from dateutil.relativedelta import relativedelta
from dateutil.rrule import (
    DAILY,
    HOURLY,
    MINUTELY,
    MONTHLY,
    SECONDLY,
    WEEKLY,
    YEARLY,
    rrule,
)
from munch import munchify

from grafanimate.grafana import GrafanaWrapper
from grafanimate.model import AnimationSequence, SequencingMode

logger = logging.getLogger(__name__)


class SequentialAnimation:
    def __init__(self, grafana: GrafanaWrapper, dashboard_uid=None, options=None):

        self.grafana = grafana
        self.dashboard_uid = dashboard_uid
        self.options = options or None

    def start(self):
        self.log("Opening dashboard")
        self.grafana.open_dashboard(self.dashboard_uid, options=self.options)
        self.log("Dashboard ready")

    def log(self, message):
        logger.info(message)
        self.grafana.console_info(message)

    def run(self, step: AnimationSequence):

        self.log("Starting animation: {}".format(step))

        # Destructure `step` instance.
        dtstart, dtuntil, interval, mode = attrgetter("dtstart", "dtuntil", "interval", "mode")(step)

        if dtstart > dtuntil:
            message = "Timestamp dtstart={} is after dtuntil={}".format(dtstart, dtuntil)
            raise ValueError(message)

        rr_freq, rr_interval, dtdelta = self.get_freq_delta(interval)

        # until = datetime.now()
        if mode == SequencingMode.CUMULATIVE:
            dtuntil += dtdelta

        # Compute complete date range.
        logger.info("Creating rrule: dtstart=%s, until=%s, freq=%s, interval=%s", dtstart, dtuntil, rr_freq, rr_interval)
        daterange = list(rrule(dtstart=dtstart, until=dtuntil, freq=rr_freq, interval=rr_interval))
        # logger.info('Date range is: %s', daterange)

        # Iterate date range.
        for date in daterange:

            logger.info("=" * 42)
            logger.info("Datetime step: %s", date)

            # Compute start and end dates based on mode.

            if mode == SequencingMode.WINDOW:
                dtstart = date
                dtuntil = date + dtdelta

            elif mode == SequencingMode.CUMULATIVE:
                dtuntil = date

            # Render image.
            image = self.render(dtstart, dtuntil, interval)

            # Build item model.
            item = munchify(
                {
                    "meta": {
                        "grafana": self.grafana,
                        "scenario": self.options["scenario"],
                        "dashboard": self.dashboard_uid,
                        "interval": interval,
                    },
                    "data": {
                        "dtstart": dtstart,
                        "dtuntil": dtuntil,
                        "image": image,
                    },
                }
            )

            yield item

            if self.options["exposure-time"] > 0:
                logger.info("Waiting for {} seconds (exposure time)".format(self.options["exposure-time"]))
                time.sleep(self.options["exposure-time"])

        self.log("Animation finished")

    def get_freq_delta(self, interval):

        rr_interval = 1

        # Secondly
        if interval == "secondly":
            rr_freq = SECONDLY
            delta = timedelta(seconds=1)

        # Minutely
        elif interval == "minutely":
            rr_freq = MINUTELY
            delta = timedelta(minutes=1) - timedelta(seconds=1)

        # Each 5 minutes
        elif interval == "5min":
            rr_freq = MINUTELY
            rr_interval = 5
            delta = timedelta(minutes=5) - timedelta(seconds=1)

        # Each 10 minutes
        elif interval == "10min":
            rr_freq = MINUTELY
            rr_interval = 10
            delta = timedelta(minutes=10) - timedelta(seconds=1)

        # Each 30 minutes
        elif interval == "30min":
            rr_freq = MINUTELY
            rr_interval = 30
            delta = timedelta(minutes=30) - timedelta(seconds=1)

        # Hourly
        elif interval == "hourly":
            rr_freq = HOURLY
            delta = timedelta(hours=1) - timedelta(seconds=1)

        # Daily
        elif interval == "daily":
            rr_freq = DAILY
            delta = timedelta(days=1) - timedelta(seconds=1)

        # Weekly
        elif interval == "weekly":
            rr_freq = WEEKLY
            delta = timedelta(weeks=1) - timedelta(seconds=1)

        # Monthly
        elif interval == "monthly":
            rr_freq = MONTHLY
            delta = relativedelta(months=+1) - relativedelta(seconds=1)

        # Yearly
        elif interval == "yearly":
            rr_freq = YEARLY
            delta = relativedelta(years=+1) - relativedelta(seconds=1)

        else:
            raise ValueError('Unknown interval "{}"'.format(interval))

        return rr_freq, rr_interval, delta

    def render(self, dtstart, dtuntil, interval):

        logger.debug("Adjusting time range control")
        self.grafana.timewarp(dtstart, dtuntil, interval)

        logger.debug("Rendering image")
        return self.make_image()

    def make_image(self):
        image = self.grafana.render_image()
        # logger.info('Image size: %s', len(image))
        return image



# -*- coding: utf-8 -*-
# (c) 2018-2021 Andreas Motl <andreas@hiveeyes.org>
# License: GNU Affero General Public License, Version 3
import logging
from pathlib import Path

import pkg_resources
from furl import furl
from munch import Munch

from grafanimate.animations import SequentialAnimation
from grafanimate.grafana import GrafanaWrapper
from grafanimate.model import AnimationScenario, AnimationSequence
from grafanimate.spool import TemporaryStorage
from grafanimate.util import as_list, filter_dict, import_module

log = logging.getLogger(__name__)


def make_grafana(url, use_panel_events) -> GrafanaWrapper:

    do_login = False
    url = furl(url)
    if url.username:
        do_login = True
        username = url.username
        password = url.password
        url.username = None
        url.password = None

    grafana = GrafanaWrapper(baseurl=str(url), use_panel_events=use_panel_events)
    grafana.boot_firefox(headless=False)
    grafana.boot_grafana()

    if do_login:
        grafana.login(username, password)

    return grafana


def get_scenario(source: str) -> AnimationScenario:
    """
    Resolve scenario from Python module or file.
    """

    # If it's not a full-qualified reference, fall back to trying the built-in scenario methods.
    modname, _, symbol = source.partition(":")
    if not symbol:
        symbol = modname
        modname = "grafanimate.scenarios"

    # Load module and resolve symbol.
    module = load_module(modname)
    scenario: AnimationScenario = resolve_reference(module, symbol)

    if scenario is None:
        raise NotImplementedError('Animation scenario "{}" not found or implemented'.format(source))

    scenario.source = source

    return scenario


def load_module(modname):
    if Path(modname).exists():
        module = import_module("<unknown>", modname)
    else:
        module = pkg_resources.EntryPoint(None, modname).resolve()
    return module


def resolve_reference(module, symbol):
    reference = getattr(module, symbol, None)
    if callable(reference):
        reference = reference()
    if isinstance(reference, AnimationScenario):
        pass
    elif isinstance(reference, (AnimationSequence, list)):
        reference = AnimationScenario(steps=as_list(reference))
    return reference


def run_animation_scenario(scenario: AnimationScenario, grafana: GrafanaWrapper, options: Munch) -> TemporaryStorage:

    log.info("Running animation scenario {}".format(scenario))

    storage = TemporaryStorage()

    # Define options to be propagated to the Javascript client domain.
    animation_options = filter_dict(
        options,
        [
            "panel-id",
            "dashboard-view",
            "header-layout",
            "datetime-format",
            "exposure-time",
            "use-panel-events",
            "scenario",
        ],
    )

    # Start the engines.
    animation = SequentialAnimation(grafana=grafana, dashboard_uid=scenario.dashboard_uid, options=animation_options)
    animation.start()

    # Run animation scenario.
    for step in scenario.steps:
        results = animation.run(step)
        storage.save_items(results)

    return storage


def run_animation_adhoc():
    # TODO: Introduce ad-hoc mode. In the meanwhile, please use scenario mode.
    """
    animator = SequentialAnimation(
        options['url'],
        time_start=options.get('start'),
        time_end=options.get('end', 'now'),
        time_step=options.get('interval', '1h')
    )
    animator.run()
    """
    pass



# -*- coding: utf-8 -*-
# (c) 2021 Andreas Motl <andreas.motl@panodata.org>
# License: GNU Affero General Public License, Version 3
import dataclasses
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union

import dateutil.parser
from dataclass_property import dataclass


class SequencingMode(Enum):
    WINDOW = "window"
    CUMULATIVE = "cumulative"


@dataclass
class AnimationSequence:
    interval: str
    mode: Optional[SequencingMode] = SequencingMode.WINDOW
    milliseconds: int = 0

    @property
    def dtstart(self) -> datetime:
        return self._dtstart

    @property
    def dtuntil(self) -> datetime:
        return self._dtuntil

    @dtstart.setter
    def dtstart(self, value: Union[datetime, str]):
        self._dtstart = self.convert_timestamp(value)

    @dtuntil.setter
    def dtuntil(self, value: Union[datetime, str]):
        self._dtuntil = self.convert_timestamp(value)

    def convert_timestamp(self, value: Union[datetime, str]) -> datetime:
        if isinstance(value, datetime):
            pass
        elif isinstance(value, int):
            value = datetime.fromtimestamp(value)
        elif isinstance(value, str):
            value = dateutil.parser.parse(value)
        else:
            raise TypeError("Unknown data type for `dtstart` or `dtuntil` value: {} ({})".format(value, type(value)))
        return value


@dataclasses.dataclass
class AnimationScenario:
    steps: List[AnimationSequence]
    grafana_url: Optional[str] = None
    dashboard_uid: Optional[str] = None
    dashboard_title: Optional[str] = None
    source: Optional[str] = None


@dataclasses.dataclass
class RenderingOptions:
    video_framerate: int = 2
    video_fps: int = 25
    gif_fps: int = 10
    gif_width: int = 480



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, "README.rst")).read()

requires = [
    # Foundation
    "munch>=2.3,<3",
    "docopt>=0.6,<1",
    "tqdm>=4,<5",
    "unidecode>=1,<2",
    "furl>=2,<3",
    "dataclass-property>=1,<2",
    # Grafana control and animation
    "where>=1,<2",
    "marionette_driver>=3,<4",
    "python-dateutil>=2.7,<3",
    "datetime-interval==0.2",
    #'Pillow==5.2.0',
]

extras = {
    "test": [
        "pytest>=6,<7",
    ],
}

setup(
    name="grafanimate",
    version="0.6.0",
    description="Animate timeseries data with Grafana",
    long_description=README,
    license="AGPL 3, EUPL 1.2",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Telecommunications Industry",
        "Topic :: Communications",
        "Topic :: Database",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Archiving",
        "Topic :: System :: Networking :: Monitoring",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    author="Andreas Motl",
    author_email="andreas.motl@panodata.org",
    url="https://github.com/panodata/grafanimate",
    keywords="grafana animate animation automation time gif video",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "grafanimate": [
            "*.js",
        ],
    },
    zip_safe=False,
    test_suite="grafanimate.test",
    install_requires=requires,
    extras_require=extras,
    tests_require=extras["test"],
    dependency_links=[],
    entry_points={
        "console_scripts": [
            "grafanimate = grafanimate.commands:run",
        ],
    },
)
